// ICE definition file for TCS v. 0.9
// written by S. Shumko
// 2005-01-19
// Modification: G. Yang 2005-02-23
// Modification: S. Shumko 2005-02-24
// Modification: S.Shumko and G.Yang 2005-03-04
// Modification: S. Shumko 2008-06-19 (change nonmutating to idempotent)

module tcs
{

	exception Error {
		string reason;
	};

	interface TCScom
	{
                // method to register a subsystem
		// name - is name of the subsystem sending this message
                void Register(string name);

                // method to unregister clients
		// name - is name of the subsystem sending this message
                void Unregister(string name);
                
		// method to send command 
		// if return value equals false then sendCommand
                // is in asynchronous mode and you should expect
		// sendReply message later, reporting on success or failure
		// of the command
		bool SendCommand(string command) throws Error;

		// reports success or failure of previously sent command 
		// if in asynchronous mode (sendCommand returned false)
		idempotent void SendReply(string reply);
               
                // method to send notification of status change etc.
                idempotent void SendNotification(string notification);

		// method to request information from the subsystem or HQ
		idempotent void RequestInformation(string request, 
			out string information) throws Error;

		// method to inform HQ or GUI of warnings, errors. 
		idempotent void SendError(string error);
	};
};